/* eslint-disable class-methods-use-this, no-console, no-alert, no-undef, arrow-body-style */
const reducers = require('./reducers');
const ConfigurationError = require('./ConfigurationError');

const Directions = {
  ASC: 'ASC',
  DESC: 'DESC',
};

/** Class representing a PivotTable. */
module.exports = class PivotTableCore {
  // public API methods

  /**
   * Initialize the pivot table.
   * @param {object} data - Row data.
   * @param {object} config - PivotTableCore configuration.
   * @param {object} metaData - Meta data.
   */
  initialize(data, config, metaData) {
    PivotTableCore.validateConfig(config);
    this.activePivots = [];
    this.base = null;
    this.config = config || {};
    this.data = data || {};
    this.filter = {
      conjunction: '&&',
      predicates: [],
    };
    this.minOrder = 1;
    this.maxOrder = 1;
    this.minOrderSubTable = 1;
    this.maxOrderSubTable = 1;
    this.metaData = metaData || {};
    this.pagination = {
      isActive: false,
      currentPage: 0,
      rowsPerPage: 15,
    };
    this.pivots = {};
    this.querySet = [];
    this.rows = {};
    this.searchString = null;
    this.sorting = {
      columnName: null,
      direction: Directions.ASC,
    };
    this.subTableColumnDefinitions = [];
    this.metaData = metaData || {};
    this.createColumnDefinitions();
    this.setInitialPivots();
    this.createRowData(this.data);
    this.pivot();
    this.applyPostProcessing('initialize', this);
  }

  /**
   * Returns an aggregate from the active pivot.
   * @param {string} key Aggregate key (i.e. it's id).
   * @returns {Object}
   */
  getAggregate(key) {
    const pivotKey = this.getActivePivotKey();
    const pivot = this.pivots[pivotKey];
    if (key in pivot) {
      return pivot[key];
    }
    return null;
  }

  /**
   * Returns all aggregates from the active pivot.
   * @returns {Object}
   */
  getAggregates() {
    const activePivotKey = this.getActivePivotKey();
    if (activePivotKey in this.pivots) {
      return this.pivots[activePivotKey];
    }
    return null;
  }

  /**
   * Returns the active pivot's columns names.
   * @returns {Array}
   */
  getActivePivotColumns() {
    return this.activePivots;
  }

  /**
   * Returns a composite key formed by the active pivot's columns names.
   * @returns {string}
   */
  getActivePivotKey() {
    return this.getActivePivotColumns().join(',');
  }

  getColumnDefinitionByName(pivotName) {
    return this.columnDefinitions.find(definition => definition.name === pivotName);
  }

  /**
   * Returns a pivot (an object containing aggregates)
   * @param {string} key Pivot key (column names separated by comma)
   * @returns {Object} A pivot.
   */
  getPivot(key) {
    if (key in this.pivots) {
      return this.pivots[key];
    }
    return null;
  }

  /**
   * Aggregates row data by active pivots, summarizes data and runs optional
   * post processing specified in config.
   */
  pivot() {
    const pivotKey = this.getActivePivotKey();
    const groupByKeys = this.createGroupByKeys();
    const aggregates = this.aggregateRows(pivotKey, groupByKeys);
    Object.keys(aggregates).forEach((key) => {
      const aggregate = aggregates[key];
      const summarizedAggregate = this.summarizeAggregate(aggregate);
      aggregates[key] = summarizedAggregate;
    });
    this.pivots[pivotKey] = aggregates;
    this.resetQuerySet();
  }

  /**
   * Returns the aggregate data filtered by the query set, with pagination info.
   */
  query() {
    return {
      pagination: this.pagination,
      rows: this.paginateQuerySet(),
    };
  }

  /*
   * Resets the query set (to unfiltered and unsorted aggregate IDs).
   */
  resetQuerySet() {
    const aggregates = this.getAggregates();
    this.querySet = Object.keys(aggregates);
  }

  /**
   * Sets the active pivot. (Doesn't perform the pivot logic though)
   * @param {Array} activePivots An array of column names.
   */
  setActivePivots(activePivots) {
    if (!Array.isArray(activePivots)) {
      throw TypeError('Argument must be an array of column names');
    }
    this.activePivots = activePivots;
  }

  /**
   * Sets filter options for the query set
   * @param {string} conjunction Determines if all or any of the predicates
   * should be matched ('&&' or '||')
   * @param {Array} predicates A list of predicate objects: column name (string)
   * , value (any), and a comparison (string). Supported comparisons:
   * '==', '!=', '<', '<=', '>', '>=', and 'in'.
   */
  setFilter(conjunction = '&&', predicates = []) {
    this.filter = { conjunction, predicates };
  }

  /**
   * Sets pagination options for the query set
   * @param {number} currentPage
   * @param {number} rowsPerPage
   */
  setPagination(isActive = true, currentPage = 0, rowsPerPage = 15) {
    this.pagination = { isActive, currentPage, rowsPerPage };
  }

  setSearchString(searchString) {
    this.searchString = searchString;
  }

  /**
    * Sets sorting options for the query set
    * @param {string} columnName Column to sort by.
    * @param {string} direction Sorting direction ('ASC' or 'DESC').
    */
  setSorting(columnName, direction = Directions.ASC) {
    this.sorting = { columnName, direction };
  }


  updateCell(rowId, columnName, value) {
    const aggregate = getAggregate(rowId);
    aggregate[`${columnName}_edit`] = value;
  }

  // private class methods

  aggregateRows(pivotKey, groupByKeys) {
    return (pivotKey in this.pivots)
      ? { ...this.pivots[pivotKey] }
      : PivotTableCore.groupBy(this.rowData, groupByKeys);
  }

  aggregateSubTable(oldAggregate) {
    if (!('subTable' in this.config)) {
      return oldAggregate;
    }
    // e.g. 'creative'
    const sKey = this.config.subTable.data;

    // copy aggregate
    const aggregate = Object.assign(oldAggregate, { [sKey]: {} });
    const sTable = aggregate[sKey];
    // group by
    aggregate.rows.forEach((rowId) => {
      // get the row
      const row = this.rowData[rowId];
      // for each sub-row
      Object.entries(row[sKey]).forEach(([sRowId, sRow]) => {
        if (sRowId in sTable) {
          // add sub-row to sub-aggregate rows
          sTable[sRowId].rows.push(sRow);
        } else {
          // create new sub-aggregate
          sTable[sRowId] = { rows: [sRow] };
        }
      });
    });
    // summarize: for each sub-aggregate
    Object.keys(sTable).forEach((sAggId) => {
      let sAgg = sTable[sAggId];
      // for each sub table column, run reducer
      this.subTableColumnDefinitions.forEach((column) => {
        sAgg[column.name] = PivotTableCore.reduce(
          sAgg.rows,
          column.name,
          column.reducer,
        );
      });

      // compute derived columns (do this iteratively,
      // for higher order derivations)
      sAgg = PivotTableCore.derive(
        sAgg,
        this.subTableColumnDefinitions,
        this.minOrderSubTable,
        this.maxOrderSubTable,
      );
      sTable[sAggId] = sAgg;
    });
    return aggregate;
  }

  applyPostProcessing(functionName, argument) {
    if ('postProcessing' in this.config
      && functionName in this.config.postProcessing
      && typeof this.config.postProcessing[functionName] === 'function') {
      this.config.postProcessing[functionName].bind(this)(argument);
    }
  }

  createColumnDefinitions() {
    // helper function
    const create = (column, subTable) => {
      // make a copy
      const definition = { ...column };
      // set name
      if (!('name' in definition)) {
        definition.name = definition.data;
      }
      // make reducer method based on column kind
      definition.reducer = PivotTableCore.makeReducerMethod(definition);
      // set pivot table base
      if (!subTable && 'base' in definition) {
        this.base = definition.name;
      }
      return definition;
    };

    if ('columns' in this.config) {
      this.columnDefinitions = this.config.columns.map((column) => {
        if ('order' in column) {
          if (column.order > this.maxOrder) {
            this.maxOrder = column.order;
          }
          if (column.order < this.minOrder) {
            this.minOrder = column.order;
          }
        }
        return create(column, false);
      });
    }

    if ('subTable' in this.config
        && 'columns' in this.config.subTable) {
      this.subTableColumnDefinitions = this.config.subTable.columns.map((column) => {
        if ('order' in column) {
          if (column.order > this.maxOrderSubTable) {
            this.maxOrderSubTable = column.order;
          }
          if (column.order < this.minOrderSubTable) {
            this.minOrderSubTable = column.order;
          }
        }
        return create(column, true);
      });
    }
  }

  createGroupByKeys() {
    return this.activePivots.map(
      activePivot => this.columnDefinitions.find(
        column => column.name === activePivot,
      ).id,
    );
  }

  createRowData(data) {
    /* Sets up pivot data using column definitions */
    const rowData = {};

    Object.keys(data).forEach((key) => {
      const dataPoint = data[key];
      const row = { id: key };
      this.columnDefinitions.forEach((columnDefinition) => {
        switch (columnDefinition.kind) {
          case 'clone':
          case 'derived':
          case 'value':
            row[columnDefinition.data] = dataPoint[columnDefinition.data];
            break;
          case 'link':
            row[columnDefinition.name] = false;
            break;
          case 'pivot':
            row[columnDefinition.data] = dataPoint[columnDefinition.data];
            row[columnDefinition.id] = ('base' in columnDefinition)
              ? key
              : dataPoint[columnDefinition.id];
            break;
          default:
            // control or anything else, do nothing
            break;
        }
      });

      // copy sub table data
      if (this.config.subTable) {
        const subKey = this.config.subTable.data;
        row[subKey] = { ...dataPoint[subKey] };
      }

      // call the post-process callback if supplied
      this.applyPostProcessing('createRowData', row);
      rowData[key] = row;
    });
    this.rowData = rowData;
  }

  filterQuerySet() {
    const aggregates = this.getAggregates();
    this.querySet = this.querySet.filter((id) => {
      // test against predicates and collect booleans
      const conditions = this.filter.predicates.map((predicate) => {
        const { value: b, comparison, column } = predicate;
        const a = aggregates[id][column];
        return PivotTableCore.compare(a, b, comparison);
      });

      const isTrue = p => p;

      switch (this.filter.conjunction) {
        case '&&':
          return conditions.every(isTrue);
        case '||':
          return conditions.some(isTrue);
        default:
          return false;
      }
    });
  }

  getRow(key) {
    if (key in this.rowData) {
      return this.rowData[key];
    }
    return null;
  }

  paginateQuerySet() {
    const { currentPage, isActive, rowsPerPage } = this.pagination;
    if (!isActive) {
      return this.querySet;
    }
    const numberOfPages = Math.ceil(this.querySet.length / rowsPerPage);
    this.pagination.numberOfPages = numberOfPages;
    const offset = rowsPerPage * currentPage;
    return this.querySet.slice(offset, offset + rowsPerPage);
  }

  setInitialPivots() {
    if ('initialPivots' in this.metaData) {
      this.activePivots = this.metaData.initialPivots;
    } else {
      this.activePivots = [this.base];
    }
  }

  searchQuerySet() {
    if (this.searchString === null || this.searchString === '') {
      return;
    }
    const searchStringLower = this.searchString.toLowerCase();
    const aggregates = this.getAggregates();
    const searchableColumns = this.columnDefinitions
      .filter(col => col.searchable)
      .map(col => col.name);
    this.querySet = this.querySet.filter((id) => {
      const aggregate = aggregates[id];
      return searchableColumns.some((col) => {
        const toSearch = aggregate[col].toLowerCase();
        return toSearch.indexOf(searchStringLower) !== -1;
      });
    });
  }

  sortQuerySet() {
    const aggregates = this.getAggregates();
    const { columnName, direction } = this.sorting;

    this.querySet.sort((id1, id2) => {
      const a = aggregates[id1][columnName];
      const b = aggregates[id2][columnName];
      if (a < b) {
        return -1;
      }
      if (a > b) {
        return 1;
      }
      return 0;
    });

    if (direction === Directions.DESC) {
      this.querySet.reverse();
    }
  }

  summarizeAggregate(_aggregate) {
    // copy aggregate
    let aggregate = { ..._aggregate };

    // get row data from ids
    const rowData = aggregate.rows.map(rowId => this.rowData[rowId]);

    // run reducers on each column
    this.columnDefinitions.forEach((column) => {
      let key = column.name;
      const editKey = `${key}_edit`;
      // if base or no reducer, copy values from first row
      // otherwise, run the reducer
      if (this.base in this.activePivots || !('reducer' in column)) {
        key = (editKey in aggregate) ? editKey : key;
        aggregate[key] = aggregate.rowData[0][key];
      } else {
        aggregate[key] = PivotTableCore.reduce(rowData, key, column.reducer);
      }
    });

    // compute derived columns (do this iteratively,
    // for higher order derivations)
    aggregate = PivotTableCore.derive(
      aggregate,
      this.columnDefinitions,
      this.minOrder,
      this.maxOrder,
    );

    this.applyPostProcessing('summarizeAggregate', aggregate);
    // return aggregate;

    aggregate = this.aggregateSubTable(aggregate);
    return aggregate;
  }

  // static methods

  static compare(a, b, comparison) {
    switch (comparison) {
      case '==':
      case '===':
        return a === b;
      case '!=':
      case '!==':
        return a !== b;
      case '>':
        return a > b;
      case '>=':
        return a >= b;
      case '<':
        return a < b;
      case '<=':
        return a <= b;
      case 'in':
        return a in b;
      default:
        return false;
    }
  }

  static derive(oldAggregate, columnDefinitions, low, high) {
    const aggregate = { ...oldAggregate };
    for (let i = low; i <= high; i += 1) {
      for (let j = 0; j < columnDefinitions.length; j += 1) {
        const column = columnDefinitions[j];
        if (column.kind === 'derived' && column.order === i) {
          aggregate[column.name] = column.derive(aggregate);
        }
      }
    }
    return aggregate;
  }

  static groupBy(rows, keys) {
    const pivot = {};
    Object.entries(rows).forEach(([rowId, row]) => {
      const aggregateKey = keys.map(key => row[key]).join(',');
      if (aggregateKey in pivot) {
        pivot[aggregateKey].rows.push(rowId);
      } else {
        pivot[aggregateKey] = { rows: [rowId] };
      }
    });
    return pivot;
  }

  static makeReducerMethod(columnDefinition) {
    switch (columnDefinition.kind) {
      case 'value':
        return (columnDefinition.reducer in reducers)
          ? reducers[columnDefinition.reducer]
          : reducers.blank;
      case 'derived':
        return reducers.sum;
      case 'pivot':
        return reducers.allEqual;
      default:
        return null;
    }
  }

  static reduce(rows, key, reducer) {
    /*
    * Applies a reduce function on a particular field for each object in a list
    * of objects. Returns a single value. If no function specified, returns
    * an empty string.
    */
    if (rows.length === 0 || reducer === null) {
      return '';
    }
    // extract values from rows
    const values = rows.map((row) => {
      // prefer to use the edited value over the original
      const editKey = `${key}_edit`;
      const value = (editKey in row) ? row[editKey] : row[key];
      return value;
    });
    return values.reduce(reducer);
  }

  static validateConfig(config) {
    // Validate columns
    function validateColumns(columns) {
      columns.forEach((column) => {
        if (!('kind' in column)) {
          throw ConfigurationError('Column properties must have a "kind" property');
        }
        switch (column.kind) {
          case 'derived':
            if (!('derive' in column)
              || typeof column.derive !== 'function') {
              throw ConfigurationError('Column properties of kind "derived" must have a "derive" function property');
            }
            if (!('order' in column)) {
              throw ConfigurationError('Derived columns must have an order property.');
            }
            break;
          case 'pivot':
            if (!('data' in column)) {
              throw ConfigurationError('Column properties of kind "pivot" must have a "data" property');
            }
            if (!('id' in column)) {
              throw ConfigurationError('Column properties of kind "pivot" must have an "id" property');
            }
            break;
          case 'value':
            if (!('data' in column)) {
              throw ConfigurationError('Column properties of kind "value" or "clone" must have a "data" property');
            }
            break;
          default:
            break;
        }
      });
    }
    if ('columns' in config) {
      validateColumns(config.columns);
      if ('subtable' in config && 'columns' in config.subtable) {
        validateColumns(config.subTable.columns);
      }
    }
  }
};
