const PivotTableCore = require('../PivotTableCore');
const data = require('./testData.json');
const config = require('./test.config');
const reducers = require('../reducers');

const metaData = { initialPivots: ['rollup_name'] };

describe('Static methods', () => {
  test('compare()', () => {
    const a = 1;
    const b = 2;
    const comparison = '<=';
    const expected = true;
    const result = PivotTableCore.compare(a, b, comparison);
    expect(result).toBe(expected);
  });

  test('derive()', () => {
    const aggregate = { a: 1, b: 2 };
    const columns = [
      {
        derive: x => x.a + x.b,
        kind: 'derived',
        name: 'a_plus_b',
        order: 1,
      },
      {
        derive: x => x.a_plus_b * x.a_plus_b,
        kind: 'derived',
        name: 'a_plus_b_squared',
        order: 2,
      },
    ];
    const result = PivotTableCore.derive(aggregate, columns, 1, 2);
    expect(result.a_plus_b).toBe(3);
    expect(result.a_plus_b_squared).toBe(9);
  });

  test('groupBy', () => {
    const rows = {
      1: { a: 'm', b: 'x' },
      2: { a: 'n', b: 'x' },
      3: { a: 'm', b: 'y' },
      4: { a: 'n', b: 'y' },
    };
    const keys = ['a', 'b'];
    const result = PivotTableCore.groupBy(rows, keys);
    const expected = {
      'm,x': { rows: ['1'] },
      'n,x': { rows: ['2'] },
      'm,y': { rows: ['3'] },
      'n,y': { rows: ['4'] },
    };
    expect(result).toEqual(expected);
  });

  describe('makeReducerMethod()', () => {
    test('column kind: value; no reducer => blank()', () => {
      const result = PivotTableCore.makeReducerMethod({ kind: 'value' });
      expect(result).toBe(reducers.blank);
    });

    test('column kind: value; with reducer => reducer()', () => {
      const result = PivotTableCore.makeReducerMethod({
        kind: 'value',
        reducer: 'mean',
      });
      expect(result).toBe(reducers.mean);
    });

    test('column kind: derived => sum()', () => {
      const result = PivotTableCore.makeReducerMethod({ kind: 'derived' });
      expect(result).toBe(reducers.sum);
    });

    test('column kind: pivot => allEqual()', () => {
      const result = PivotTableCore.makeReducerMethod({ kind: 'pivot' });
      expect(result).toBe(reducers.allEqual);
    });

    test('column kind: anything else => null', () => {
      const result = PivotTableCore.makeReducerMethod({ kind: 'control' });
      expect(result).toBeNull();
    });
  });

  describe('reduce()', () => {
    test('simple sum', () => {
      const rows = [
        { a: 1, b: 'x' },
        { a: 1, b: 'x' },
      ];
      const result = PivotTableCore.reduce(rows, 'a', reducers.sum);
      expect(result).toBe(2);
    });
    test('no reducer specified', () => {
      const rows = [
        { a: 1, b: 'x' },
        { a: 1, b: 'x' },
      ];
      const result = PivotTableCore.reduce(rows, 'a', null);
      expect(result).toBe('');
    });
  });
});

describe('Setup', () => {
  const pivotTable = new PivotTableCore();
  test('instantiation', () => {
    expect(pivotTable).toBeDefined();
  });

  describe('Initializing', () => {
    pivotTable.initialize(data, config, metaData);
    test('createColumnDefinitions()', () => {
      expect(pivotTable.columnDefinitions).toHaveLength(8);
    });

    test('setInitialPivots()', () => {
      expect(pivotTable.activePivots).toBe(metaData.initialPivots);
    });

    describe('createRowData()', () => {
      test('copied right amount of data', () => {
        expect(Object.keys(pivotTable.rowData)).toEqual(['1', '2', '3', '4']);
        Object.values(pivotTable.rowData, (row) => {
          expect(row).toHaveProperty('id');
          if ('creative' in row) {
            expect(row.creative).toHaveProperty('mct_name');
            expect(row.creative).toHaveProperty('opt_spend');
          }
        });
      });

      test('applyPostProcessing()', () => {
        expect(pivotTable.rowData[1].postProcessing).toBe(true);
      });
    });

    test('applyPostProcessing()()', () => {
      expect(pivotTable.postProcessing).toBe(true);
    });
  });
});

describe('Pivoting', () => {
  const pivotTable = new PivotTableCore();
  pivotTable.initialize(data, config, metaData);
  describe('Aggregating', () => {
    test('setActivePivots()', () => {
      const activePivots = ['dma_name'];
      pivotTable.setActivePivots(activePivots);
      expect(pivotTable.activePivots).toEqual(activePivots);
    });

    test('createGroupByKeys()', () => {
      pivotTable.setActivePivots(['rollup_name']);
      const result = pivotTable.createGroupByKeys();
      expect(result).toEqual(['rollup_id']);
    });
  });

  describe('Summarizing', () => {
    pivotTable.initialize(data, config, metaData);
    test('aggregate', () => {
      const aggregate = pivotTable.getAggregate('17');
      const result = pivotTable.summarizeAggregate(aggregate);
      expect(result.geography_name).toBe('National');
      expect(result.name).toBe('Mixed');
      expect(result.opt_spend).toBe(140);
      expect(result.plan_spend).toBe(160);
      expect(result.rollup_name).toBe('Network TV');
      expect(result.override_impact).toBe(150);
      expect(result.spend_difference).toBe(-20);
      expect(result.spend_difference_per_impact).toBe(-20 / 150);
    });

    test('sub table', () => {
      const aggregate = pivotTable.getAggregate('51');
      const result = pivotTable.summarizeAggregate(aggregate);
      expect(result.creative['704'].mct_name).toBe('color');
      expect(result.creative['704'].opt_spend).toBe(20);
      expect(result.creative['704'].mct_name_opt_spend).toBe('color_20');
    });
  });

  test('applyPostProcessing()', () => {
    const aggregate = pivotTable.getAggregate('17');
    const result = pivotTable.summarizeAggregate(aggregate);
    expect(result.postProcessing).toBe(true);
  });
});

describe('Querying', () => {
  const pivotTable = new PivotTableCore();
  pivotTable.initialize(data, config, metaData);
  pivotTable.setActivePivots(['name']);
  pivotTable.pivot();

  test('resetQuerySet()', () => {
    const expected = ['1', '2', '3', '4'];
    pivotTable.resetQuerySet();
    expect(pivotTable.querySet).toEqual(expected);
  });

  test('setFilter()', () => {
    const filter = {
      conjunction: '&&',
      predicates: [{ column: 'plan_spend', value: 75, comparison: '<' }],
    };
    pivotTable.setFilter(filter.conjunction, filter.predicates);
    expect(pivotTable.filter).toEqual(filter);
  });

  test('setPagination()', () => {
    const pagination = {
      isActive: true,
      currentPage: 0,
      rowsPerPage: 5,
    };
    pivotTable.setPagination(
      pagination.isActive,
      pagination.currentPage,
      pagination.rowsPerPage,
    );
    expect(pivotTable.pagination).toEqual(pagination);
  });

  test('setSearchString()', () => {
    const searchString = 'Network';
    pivotTable.setSearchString(searchString);
    expect(pivotTable.searchString).toBe(searchString);
  });

  test('setSorting()', () => {
    const sorting = {
      columnName: 'plan_spend',
      direction: 'ASC',
    };
    pivotTable.setSorting(sorting.columnName, sorting.direction);
    expect(pivotTable.sorting).toEqual(sorting);
  });

  test('filterQuerySet()', () => {
    pivotTable.resetQuerySet();
    pivotTable.filterQuerySet();
    const result = pivotTable.querySet;
    const expected = ['2', '3', '4'];
    expect(result).toEqual(expected);
  });

  test('searchQuerySet()', () => {
    pivotTable.resetQuerySet();
    pivotTable.searchQuerySet();
    const result = pivotTable.querySet;
    const expected = ['1', '2'];
    expect(result).toEqual(expected);
  });

  test('sortQuerySet()', () => {
    pivotTable.resetQuerySet();
    pivotTable.sortQuerySet();
    const result = pivotTable.querySet;
    const expected = ['3', '4', '2', '1'];
    expect(result).toEqual(expected);
  });

  test('paginateQuerySet', () => {
    pivotTable.resetQuerySet();
    pivotTable.setPagination(true, 0, 3);
    const result = pivotTable.paginateQuerySet();
    const expected = ['1', '2', '3'];
    expect(result).toEqual(expected);
  });

  test('query()', () => {
    const result = pivotTable.query();
    expect(result).toHaveProperty('pagination');
    expect(result).toHaveProperty('rows');
  });
});

describe('Getters', () => {
  const pivotTable = new PivotTableCore();
  pivotTable.initialize(data, config, metaData);
  test('getActivePivotColumns()', () => {
    const activePivots = ['rollup_name', 'geography_name'];
    pivotTable.setActivePivots(activePivots);
    const result = pivotTable.getActivePivotColumns();
    expect(result).toEqual(activePivots);
  });

  test('getActivePivotKey()', () => {
    const activePivots = ['rollup_name', 'geography_name'];
    pivotTable.setActivePivots(activePivots);
    const result = pivotTable.getActivePivotKey();
    expect(result).toBe('rollup_name,geography_name');
  });

  test('getAggregate()', () => {
    pivotTable.setActivePivots(['rollup_name']);
    pivotTable.pivot();
    const result = pivotTable.getAggregate('17');
    expect(result).toHaveProperty('rollup_name');
  });

  test('getAggregates()', () => {
    pivotTable.setActivePivots(['rollup_name']);
    pivotTable.pivot();
    const result = pivotTable.getAggregates();
    expect(result).toHaveProperty('17');
  });

  test('getColumnDefinitionByName()', () => {
    const name = 'rollup_name';
    const result = pivotTable.getColumnDefinitionByName(name);
    expect(result.name).toBe(name);
  });

  test('getPivot()', () => {
    pivotTable.setActivePivots(['rollup_name']);
    pivotTable.pivot();
    const key = 'rollup_name';
    const result = pivotTable.getPivot(key);
    expect(result).not.toBeNull();
  });

  test('getRow()', () => {
    const key = '1';
    const result = pivotTable.getRow(key);
    expect(result).not.toBeNull();
  });
});

describe('Setters', () => {
  const pivotTable = new PivotTableCore();
  pivotTable.initialize(data, config, metaData);
  test('setActivePivots()', () => {
    const activePivots = ['rollup_name'];
    pivotTable.setActivePivots(activePivots);
    expect(pivotTable.activePivots).toEqual(activePivots);
  });
  test('setInitialPivots()', () => {
    pivotTable.setInitialPivots();
    expect(pivotTable.activePivots).toEqual(['rollup_name']);
  });
});
